@@ -12,24 +12,38 @@ the terms of the GNU LGPL.
 import errno
 import functools
 import logging
-import os
+from typing import TYPE_CHECKING, Callable, NewType, Optional, Sequence, Tuple
 
-# Will be injected by pyfuse3 extension module
-FUSEError = None
+# These types are specific instances of builtin types:
+FileHandleT = NewType("FileHandleT", int)
+FileNameT = NewType("FileNameT", bytes)
+FlagT = NewType("FlagT", int)
+InodeT = NewType("InodeT", int)
+ModeT = NewType("ModeT", int)
 
-__all__ = [ 'Operations', 'async_wrapper' ]
+if TYPE_CHECKING:
+    # These types are defined elsewhere in the C code
+    from pyfuse3 import (EntryAttributes, FileInfo, FUSEError, ReaddirToken,
+                         RequestContext, SetattrFields, StatvfsData)
+else:
+    # Will be injected by pyfuse3 extension module
+    FUSEError = None
+
+__all__ = ['Operations', 'async_wrapper']
 
 log = logging.getLogger(__name__)
 
+
 # Any top level trio coroutines (i.e., coroutines that are passed
 # to trio.run) must be pure-Python. This wrapper ensures that this
 # is the case for Cython-defined async functions.
-def async_wrapper(fn):
+def async_wrapper(fn: Callable) -> Callable:
     @functools.wraps(fn)
-    async def wrapper(*args, **kwargs):
+    async def wrapper(*args, **kwargs):  # type: ignore
         await fn(*args, **kwargs)
     return wrapper
 
+
 class Operations:
     '''
     This class defines the request handler methods that an pyfuse3 file system
@@ -47,12 +61,12 @@ class Operations:
     this class all just raise ``FUSEError(ENOSYS)`` or do nothing).
     '''
 
-    supports_dot_lookup = True
-    enable_writeback_cache = False
-    enable_acl = False
+    supports_dot_lookup: bool = True
+    enable_writeback_cache: bool = False
+    enable_acl: bool = False
 
-    def init(self):
-        '''Initialize operations
+    def init(self) -> None:
+        '''Initialize operations.
 
         This method will be called just before the file system starts handling
         requests. It must not raise any exceptions (not even `FUSEError`), since
@@ -61,7 +75,12 @@ class Operations:
 
         pass
 
-    async def lookup(self, parent_inode, name, ctx):
+    async def lookup(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
         '''Look up a directory entry by name and get its attributes.
 
         This method should return an `EntryAttributes` instance for the
@@ -84,8 +103,11 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def forget(self, inode_list):
-        '''Decrease lookup counts for inodes in *inode_list*
+    async def forget(
+        self,
+        inode_list: Sequence[Tuple[InodeT, int]]
+    ) -> None:
+        '''Decrease lookup counts for inodes in *inode_list*.
 
         *inode_list* is a list of ``(inode, nlookup)`` tuples. This method
         should reduce the lookup count for each *inode* by *nlookup*.
@@ -107,8 +129,12 @@ class Operations:
 
         pass
 
-    async def getattr(self, inode, ctx):
-        '''Get attributes for *inode*
+    async def getattr(
+        self,
+        inode: InodeT,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
+        '''Get attributes for *inode*.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -119,9 +145,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-
-    async def setattr(self, inode, attr, fields, fh, ctx):
-        '''Change attributes of *inode*
+    async def setattr(
+        self,
+        inode: InodeT,
+        attr: "EntryAttributes",
+        fields: "SetattrFields",
+        fh: Optional[FileHandleT],
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
+        '''Change attributes of *inode*.
 
         *fields* will be an `SetattrFields` instance that specifies which
         attributes are to be updated. *attr* will be an `EntryAttributes`
@@ -147,7 +179,11 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def readlink(self, inode, ctx):
+    async def readlink(
+        self,
+        inode: InodeT,
+        ctx: "RequestContext"
+    ) -> FileNameT:
         '''Return target of symbolic link *inode*.
 
         *ctx* will be a `RequestContext` instance.
@@ -155,9 +191,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-
-    async def mknod(self, parent_inode, name, mode, rdev, ctx):
-        '''Create (possibly special) file
+    async def mknod(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        mode: ModeT,
+        rdev: int,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
+        '''Create (possibly special) file.
 
         This method must create a (special or regular) file *name* in the
         directory with inode *parent_inode*. Whether the file is special or
@@ -170,12 +212,18 @@ class Operations:
 
         (Successful) execution of this handler increases the lookup count for
         the returned inode by one.
-       '''
+        '''
 
         raise FUSEError(errno.ENOSYS)
 
-    async def mkdir(self, parent_inode, name, mode, ctx):
-        '''Create a directory
+    async def mkdir(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        mode: ModeT,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
+        '''Create a directory.
 
         This method must create a new directory *name* with mode *mode* in the
         directory with inode *parent_inode*. *ctx* will be a `RequestContext`
@@ -190,8 +238,13 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def unlink(self, parent_inode, name, ctx):
-        '''Remove a (possibly special) file
+    async def unlink(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        ctx: "RequestContext"
+    ) -> None:
+        '''Remove a (possibly special) file.
 
         This method must remove the (special or regular) file *name* from the
         direcory with inode *parent_inode*.  *ctx* will be a `RequestContext`
@@ -207,13 +260,17 @@ class Operations:
         the `forget` method to be carried out when the lookup count reaches zero
         (and of course only if at that point there are no more directory entries
         associated with the inode either).
-
         '''
 
         raise FUSEError(errno.ENOSYS)
 
-    async def rmdir(self, parent_inode, name, ctx):
-        '''Remove directory *name*
+    async def rmdir(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        ctx: "RequestContext"
+    ) -> None:
+        '''Remove directory *name*.
 
         This method must remove the directory *name* from the direcory with
         inode *parent_inode*. *ctx* will be a `RequestContext` instance. If
@@ -236,8 +293,14 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def symlink(self, parent_inode, name, target, ctx):
-        '''Create a symbolic link
+    async def symlink(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        target: FileNameT,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
+        '''Create a symbolic link.
 
         This method must create a symbolink link named *name* in the directory
         with inode *parent_inode*, pointing to *target*.  *ctx* will be a
@@ -252,8 +315,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def rename(self, parent_inode_old, name_old, parent_inode_new,
-               name_new, flags, ctx):
+    async def rename(
+        self,
+        parent_inode_old: InodeT,
+        name_old: str,
+        parent_inode_new: InodeT,
+        name_new: str,
+        flags: FlagT,
+        ctx: "RequestContext"
+    ) -> None:
         '''Rename a directory entry.
 
         This method must rename *name_old* in the directory with inode
@@ -285,7 +355,13 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def link(self, inode, new_parent_inode, new_name, ctx):
+    async def link(
+        self,
+        inode: InodeT,
+        new_parent_inode: InodeT,
+        new_name: FileNameT,
+        ctx: "RequestContext"
+    ) -> "EntryAttributes":
         '''Create directory entry *name* in *parent_inode* refering to *inode*.
 
         *ctx* will be a `RequestContext` instance.
@@ -299,7 +375,12 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def open(self, inode, flags, ctx):
+    async def open(
+        self,
+        inode: InodeT,
+        flags: FlagT,
+        ctx: "RequestContext"
+    ) -> "FileInfo":
         '''Open a inode *inode* with *flags*.
 
         *ctx* will be a `RequestContext` instance.
@@ -317,10 +398,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def read(self, fh, off, size):
-        '''Read *size* bytes from *fh* at position *off*
+    async def read(
+        self,
+        fh: FileHandleT,
+        off: int,
+        size: int
+    ) -> int:
+        '''Read *size* bytes from *fh* at position *off*.
 
-        *fh* will by an integer filehandle returned by a prior `open` or
+        *fh* will be an integer filehandle returned by a prior `open` or
         `create` call.
 
         This function should return exactly the number of bytes requested except
@@ -330,10 +416,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def write(self, fh, off, buf):
-        '''Write *buf* into *fh* at *off*
+    async def write(
+        self,
+        fh: FileHandleT,
+        off: int,
+        buf: bytes
+    ) -> int:
+        '''Write *buf* into *fh* at *off*.
 
-        *fh* will by an integer filehandle returned by a prior `open` or
+        *fh* will be an integer filehandle returned by a prior `open` or
         `create` call.
 
         This method must return the number of bytes written. However, unless the
@@ -344,10 +435,13 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def flush(self, fh):
+    async def flush(
+        self,
+        fh: FileHandleT
+    ) -> None:
         '''Handle close() syscall.
 
-        *fh* will by an integer filehandle returned by a prior `open` or
+        *fh* will be an integer filehandle returned by a prior `open` or
         `create` call.
 
         This method is called whenever a file descriptor is closed. It may be
@@ -357,14 +451,17 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def release(self, fh):
-        '''Release open file
+    async def release(
+        self,
+        fh: FileHandleT
+    ) -> None:
+        '''Release open file.
 
         This method will be called when the last file descriptor of *fh* has
         been closed, i.e. when the file is no longer opened by any client
         process.
 
-        *fh* will by an integer filehandle returned by a prior `open` or
+        *fh* will be an integer filehandle returned by a prior `open` or
         `create` call. Once `release` has been called, no future requests for
         *fh* will be received (until the value is re-used in the return value of
         another `open` or `create` call).
@@ -375,20 +472,28 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def fsync(self, fh, datasync):
-        '''Flush buffers for open file *fh*
+    async def fsync(
+        self,
+        fh: FileHandleT,
+        datasync: bool
+    ) -> None:
+        '''Flush buffers for open file *fh*.
 
         If *datasync* is true, only the file contents should be
         flushed (in contrast to the metadata about the file).
 
-        *fh* will by an integer filehandle returned by a prior `open` or
+        *fh* will be an integer filehandle returned by a prior `open` or
         `create` call.
         '''
 
         raise FUSEError(errno.ENOSYS)
 
-    async def opendir(self, inode, ctx):
-        '''Open the directory with inode *inode*
+    async def opendir(
+        self,
+        inode: InodeT,
+        ctx: "RequestContext"
+    ) -> FileHandleT:
+        '''Open the directory with inode *inode*.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -399,8 +504,12 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-
-    async def readdir(self, fh, start_id, token):
+    async def readdir(
+        self,
+        fh: FileHandleT,
+        start_id: int,
+        token: "ReaddirToken"
+    ) -> None:
         '''Read entries in open directory *fh*.
 
         This method should list the contents of directory *fh* (as returned by a
@@ -431,8 +540,11 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def releasedir(self, fh):
-        '''Release open directory
+    async def releasedir(
+        self,
+        fh: FileHandleT
+    ) -> None:
+        '''Release open directory.
 
         This method will be called exactly once for each `opendir` call. After
         *fh* has been released, no further `readdir` requests will be received
@@ -441,8 +553,12 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def fsyncdir(self, fh, datasync):
-        '''Flush buffers for open directory *fh*
+    async def fsyncdir(
+        self,
+        fh: FileHandleT,
+        datasync: bool
+    ) -> None:
+        '''Flush buffers for open directory *fh*.
 
         If *datasync* is true, only the directory contents should be
         flushed (in contrast to metadata about the directory itself).
@@ -450,8 +566,11 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def statfs(self, ctx):
-        '''Get file system statistics
+    async def statfs(
+        self,
+        ctx: "RequestContext"
+    ) -> "StatvfsData":
+        '''Get file system statistics.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -460,8 +579,8 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    def stacktrace(self):
-        '''Asynchronous debugging
+    def stacktrace(self) -> None:
+        '''Asynchronous debugging.
 
         This method will be called when the ``fuse_stacktrace`` extended
         attribute is set on the mountpoint. The default implementation logs the
@@ -471,18 +590,25 @@ class Operations:
 
         import sys
         import traceback
+        from os.path import basename
 
         code = list()
         for threadId, frame in sys._current_frames().items():
-            code.append("\n# ThreadID: %s" % threadId)
+            code.append(f"\n# ThreadID: {threadId}")
             for filename, lineno, name, line in traceback.extract_stack(frame):
-                code.append('%s:%d, in %s' % (os.path.basename(filename), lineno, name))
+                code.append(f'{basename(filename)}:{lineno}, in {name}')
                 if line:
-                    code.append("    %s" % (line.strip()))
+                    code.append(f"    {line.strip()}")
 
         log.error("\n".join(code))
 
-    async def setxattr(self, inode, name, value, ctx):
+    async def setxattr(
+        self,
+        inode: InodeT,
+        name: bytes,
+        value: bytes,
+        ctx: "RequestContext"
+    ) -> None:
         '''Set extended attribute *name* of *inode* to *value*.
 
         *ctx* will be a `RequestContext` instance.
@@ -494,8 +620,13 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def getxattr(self, inode, name, ctx):
-        '''Return extended attribute *name* of *inode*
+    async def getxattr(
+        self,
+        inode: InodeT,
+        name: bytes,
+        ctx: "RequestContext"
+    ) -> bytes:
+        '''Return extended attribute *name* of *inode*.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -506,8 +637,12 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def listxattr(self, inode, ctx):
-        '''Get list of extended attributes for *inode*
+    async def listxattr(
+        self,
+        inode: InodeT,
+        ctx: "RequestContext"
+    ) -> Sequence[bytes]:
+        '''Get list of extended attributes for *inode*.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -517,8 +652,13 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def removexattr(self, inode, name, ctx):
-        '''Remove extended attribute *name* of *inode*
+    async def removexattr(
+        self,
+        inode: InodeT,
+        name: bytes,
+        ctx: "RequestContext"
+    ) -> None:
+        '''Remove extended attribute *name* of *inode*.
 
         *ctx* will be a `RequestContext` instance.
 
@@ -529,8 +669,12 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-
-    async def access(self, inode, mode, ctx):
+    async def access(
+        self,
+        inode: InodeT,
+        mode: ModeT,
+        ctx: "RequestContext"
+    ) -> bool:
         '''Check if requesting process has *mode* rights on *inode*.
 
         *ctx* will be a `RequestContext` instance.
@@ -546,8 +690,15 @@ class Operations:
 
         raise FUSEError(errno.ENOSYS)
 
-    async def create(self, parent_inode, name, mode, flags, ctx):
-        '''Create a file with permissions *mode* and open it with *flags*
+    async def create(
+        self,
+        parent_inode: InodeT,
+        name: FileNameT,
+        mode: ModeT,
+        flags: FlagT,
+        ctx: "RequestContext"
+    ) -> Tuple["FileInfo", "EntryAttributes"]:
+        '''Create a file with permissions *mode* and open it with *flags*.
 
         *ctx* will be a `RequestContext` instance.
 
